import numpy as np
import scipy as sp
from graph import Graph

# All algorithms assume that the relevant graph matrices are already computed 

# --- Graph Edit Distance ---

def graph_edit_distance(g1, g2):
    'The sum of the number of edges that are present only in g1 or g2'

    m1 = g1.get_adjacency_matrix()
    m2 = g2.get_adjacency_matrix()

    diff = m1 - m2

    return sum(sum(abs(diff)))

# --- DeltaCon ---

def root_ed(s1, s2):

    root_ed = np.sqrt(sum(sum((np.sqrt(s1) - np.sqrt(s2)) ** 2)))

    return root_ed


def fabp(g):

    adj = g.get_adjacency_matrix()
    diag = g.get_diagonal_matrix()
    identity = np.identity(adj.shape[0])
    
    e = 1/(1 + diag.max())
    s = np.linalg.inv(identity + (e ** 2) * diag - e * adj)
    
    return s


def delta_con(g1, g2):

    s1 = fabp(g1)
    s2 = fabp(g2)

    dist = root_ed(s1, s2)

    return dist

# --- SimRank Distance ---

def sim_rank(g, c = 0.8, max_iters = 10):

    adj = g.get_adjacency_matrix()
    col_degree = adj.sum(axis = 0)
    col_degree = [x if x != 0 else 1 for x in col_degree]
    col_norm_adj = adj / col_degree

    s = np.identity(adj.shape[0])
    
    for i in range(max_iters):

        s = c * np.dot(np.dot(col_norm_adj.T, s), col_norm_adj)
        np.fill_diagonal(s, np.ones(adj.shape[0]))

    return s

def sim_rank_distance(g1, g2, c = 0.8, max_iters = 10):

    s1 = sim_rank(g1, c, max_iters)
    s2 = sim_rank(g2, c, max_iters)

    dist = matrix_ed(s1, s2)

    return dist

# --- Node In/Out Degree Distance ---

def node_degree_matrix(g):

    adj = g.get_adjacency_matrix()
    in_deg = adj.sum(axis=1)
    out_deg = adj.sum(axis=0)
    s = np.vstack((in_deg, out_deg))

    return s.T

def degree_dist(g1, g2):

    s1 = node_degree_matrix(g1)
    s2 = node_degree_matrix(g2)

    dist = matrix_ed(s1, s2)

    return dist

# --- Matrix Euclidean Distance ---

def matrix_ed(m1, m2):

    dist = np.sum(np.sqrt(np.sum((m1 - m2) ** 2, axis = 1)))

    return dist

def flat_matrix_ed(m1, m2):

    dist = np.linalg.norm(m1 - m2)

    return dist


# --- Matrix Cosine Distance ---

def matrix_cd(m1, m2):

    norm1 = np.sqrt(np.sum(m1 ** 2, axis = 1))
    norm2 = np.sqrt(np.sum(m2 ** 2, axis = 1))
    norm = norm1 * norm2

    dot = np.sum(m1 * m2, axis = 1)

    dist = np.sum(np.ones(dot.shape) - dot / norm)

    return dist

def flat_matrix_cd(m1, m2):

    dist = sp.spatial.distance.cosine(m1.flatten(), m2.flatten())

    return dist

# -- Matrix Manhattan Distance --

def matrix_md(m1, m2):

    dist = sp.spatial.distance.cityblock(m1.flatten(), m2.flatten())

    return dist

# Graph Distance Algorithms Considering Node Weights

# --- Graph Edit Distance With Node Weights ---

def graph_edit_distance_nw(g1, g2):

    m1 = g1.get_adjacency_matrix()
    m2 = g2.get_adjacency_matrix()

    nw1 = g1.get_vertex_weights()
    nw2 = g2.get_vertex_weights()

    ged_diff = sum(sum(abs(m1 - m2)))
    nw_diff = sum(abs(nw1 - nw2))

    return ged_diff + nw_diff

# --- DeltaCon With Node Weights ---

def fabp_nw(g, norm):

    adj = g.get_adjacency_matrix()
    diag = g.get_diagonal_matrix()
    node_weights = g.get_vertex_weights()
    identity = np.identity(adj.shape[0])
    
    e = 1/(1 + diag.max())
    s = np.linalg.inv(identity + (e ** 2) * diag - e * adj)

    nw_matrix = np.zeros(s.shape)
    np.fill_diagonal(nw_matrix, node_weights)
    s = np.dot(nw_matrix, s) / norm
    
    return s

def get_largest_node_weight(graphs):

    return max([graph.get_vertex_weights().max() for graph in graphs])

# --- Node In/Out Degree and Node Weigh Distance ---

def node_degree_weight_matrix(g):

    adj = g.get_adjacency_matrix()
    nw = g.get_vertex_weights()
    in_deg = adj.sum(axis=1)
    out_deg = adj.sum(axis=0)
    s = np.vstack((in_deg, out_deg, nw))

    return s.T

def degree_dist_nw(g1, g2):

    s1 = node_degree_weight_matrix(g1)
    s2 = node_degree_weight_matrix(g2)

    dist = matrix_ed(s1, s2)

    return dist



import numpy as np


def __get_num_of_edges(num_of_nodes, directed, clique_range):
    'The number of edges will be equal to a percentage of the edges in a clique'

    edges_in_clique = (num_of_nodes * (num_of_nodes - 1)) / 2
    percentage = (np.random.randint(clique_range[0], clique_range[1]) + np.random.random()) / 100.0
    num_of_edges = int(round(edges_in_clique * percentage))

    # if the edges are directed we can have twice as many
    if directed:
        num_of_edges = num_of_edges * 2

    return num_of_edges


def __generate_edge_weight(weighted, weight_range):
    'Generate an edge weight'

    # if the edge is unweighted assign a weight of one, otherwise assign a random weight in the given range
    if not weighted:
        weight = 1
    else:
        weight = np.random.randint(weight_range[0], weight_range[1] + 1)

    return weight


def __generate_edges(num_of_nodes, clique_range, weighted, weight_range, directed, dense_regions, sparse_regions):
    'Generate all edges for a given graph'

    num_of_edges =  __get_num_of_edges(num_of_nodes, directed, clique_range)
    edges = []

    for e in range(num_of_edges):

        approved = False

        while not approved:

            # generate an edge in one of the dense regions
            if len(dense_regions) > 0 and np.random.random() < 0.50:
                region = dense_regions[np.random.randint(0, len(dense_regions))]
                node1 = np.random.randint(region[0], region[1])
                node2 = np.random.randint(region[0], region[1])
                edge = [node1, node2]

            # generate an edge in the rest of the graphs
            else:
                node1 = np.random.randint(0, num_of_nodes)
                node2 = np.random.randint(0, num_of_nodes)
                edge = [node1, node2]

                # if the edge is from one of the sparse regions possibly reject it
                for region in sparse_regions:
                    if edge[0] in range(region[0], region[1]) and edge[1] in range(region[0], region[1]):
                        if np.random.random() < 0.80:
                            edge = [0, 0]
                            break

            if not directed:
                edge.sort()

            # currently excluding self edges
            if edge[0] != edge[1] and edge not in edges:
                edges.append(edge)
                approved = True

    # include a weight for each edge
    for edge in edges:
        weight = __generate_edge_weight(weighted, weight_range)
        edge.append(weight)

    return edges


def generate_graphs_file(file_name, num_of_graphs, num_of_nodes, seed = 0,
                         clique_range = [30, 60], weighted = False, weight_range = [1, 10],
                         directed = False, dense_regions = [], sparse_regions = []):

    """Generate a collection of graphs with a set of dense and sparse regions

    Parameters
    ----------
    file_name: string
        name of the generated graphs file

    num_of_graphs: int
        the nuber of graphs that will be generated and written into the file

    num_of_nodes: int
        the number of nodes that every graph will have

    seed: int
        a seed paramether for controlling randomness

    clique_range: [lwb: int, upb: int], lwb and upb are in the range 0 to 100
        a value between lwb and upb will be chosen to decide how many edges a graph
        will have - calculated as a percentage of the edges in a clique

    weighted: boolean
        determines if the edges will be weighted or not

    weight_range: [lwb: int, upb: int]
        the weight of each edge will be a random integer form lwb to upb

    directed: boolean
        determines if the edges will be directed or not

    dense_regions: list of (region_start_node: int, region_end_node: int) tuples
        outlines a colleaction of densly connected regions

    sparse_regions: list of (region_start_node: int, region_end_node: int) tuples
        outlines a colleaction of sparsely connected regions
    """
    
    np.random.seed(seed)
    file = open(file_name, 'w')

    for graph in range(num_of_graphs):

        file.write(str(num_of_nodes) + '\n')

        edges = __generate_edges(num_of_nodes, clique_range, weighted, weight_range,
                                directed, dense_regions, sparse_regions)

        for edge in edges:
            file.write(str(edge[0]) + ' ' + str(edge[1]) + ' ' + str(edge[2]) + '\n')

    file.close()



from math import exp


def rbf_kernel(item1, item2, dist_func, sigma = 0.1):
    'Radial basis function - Gaussian Kernel'

    distance = dist_func(item1, item2)
    kernel = rbf(distance, sigma)

    return kernel

def rbf(distance, sigma = 0.1):
    'Radial basis function with pre-computed distance'

    square_dist = distance ** 2
    gaussian_sim = exp(-(square_dist / 2 * sigma ** 2))

    return gaussian_sim


def dc_kernel(item1, item2, dist_func):
    'Distance to similarity conversion proposed by DeltaCon'

    distance = dist_func(item1, item2)
    kernel = 1 / (1 + distance)

    return kernel



import numpy as np
from copy import deepcopy
from kmeans_init import get_random_centroids, kmeans_pp


def __get_item_indices_per_cluster(k, labels):
    'Get a 2-D array where the array at position i holds the indices of all items in cluster i'

    indices_per_cluster = [[] for x in range(k)]
    for i in range(len(labels)):
        indices_per_cluster[labels[i]].append(i)

    return indices_per_cluster


def __distance_to_centroid(item_index, centroid_index, item_indices_per_cluster, kernel_matrix):
    'Compute the distance between a given item and centroid in the projected space'

    item_indices_in_cluster = item_indices_per_cluster[centroid_index]
    num_items_in_cluster = len(item_indices_in_cluster)

    param1 = kernel_matrix[item_index, item_index]
    param2 = 0
    param3 = 0

    # check in case there are no items assigned to the cluster
    if num_items_in_cluster != 0:

        kernel_sum = 0
        for i in item_indices_in_cluster:
            kernel_sum += kernel_matrix[item_index, i]

        param2 = (-2 * kernel_sum) / num_items_in_cluster

        kernel_sum = 0
        for i in item_indices_in_cluster:
            for j in item_indices_in_cluster:
                kernel_sum += kernel_matrix[i, j]

        param3 = kernel_sum / (num_items_in_cluster ** 2)

    distance = param1 + param2 + param3

    return distance


def __random_cluster_assignment(k, seed, items):
    'Assign random cluster labels to each element'

    np.random.seed(seed)
    labels = [np.random.randint(0, k) for x in range(len(items))]
    return labels


def __proximity_cluster_assignment(k, seed, items, dist_func, init):
    'Choose K centroid items and assign labels to all items based on their proximity to the centroids'

    if init == "proxy":
        centroid_indices = get_random_centroids(k, seed, items)
    else:
        centroid_indices = kmeans_pp(k, seed, items, dist_func)

    labels = []

    for item in items:

        min_distance = float("inf")
        closest_centroid = -1

        for label in range(len(centroid_indices)):

            dist = dist_func(item, items[centroid_indices[label]])

            if dist < min_distance:
                min_distance = dist
                closest_centroid = label

        labels.append(closest_centroid)

    return labels


def get_kernel_matrix(items, dist_func, kernel, **kernel_args):
    'Compute the NxN matrix of kernel values for all pairs of items'

    kernel_matrix = np.zeros((len(items), len(items)))

    j = 0
    for i in range(len(items)):

        while j < len(items):

            kernel_res = kernel(items[i], items[j], dist_func, **kernel_args)
            kernel_matrix[i, j] = kernel_res
            kernel_matrix[j, i] = kernel_res

            j += 1

        j = i + 1

    return kernel_matrix


def kernel_kmeans(k, max_iters, seed, items, dist_func, kernel, init = "random", kernel_matrix = None, **kernel_args):
    '''Run Kernel K-Means - returns

    Parameters
    ----------
    k: int
        the number of clusters

    max_iters: int
        the number of maximum allowed itterations before the method terminates
        (termination will occure earlier if the cluster labels converge)

    seed: int
        a seed paramether for controlling randomness

    items: 1-D list
        the list of structured data items thet will be clustered

    dist_func: function pointer
        a function for computing the distance between any two items in the items list

    kernel: function pointer
        a kernel function

    init: string
        a string indicating the cluster label initialization method

    kernel_matrix: 2-D numpy array
        an optional pre-computed NxN kernal matrix

    The possible initialization methods are: 'random', 'proxy' and 'kpp':
        random - provide a random cluster label to each item
        proxy  - randomly choose representative items and assign labels based on proximity to representatives
        kpp    - similar to 'proxy' except the representatives are chosen using k-means++

    **kernel_args: optional kernel arguments
        any further parameters are passed directly to the kernel function

    Return
    ------
    labels: 1-D list of ints
        a list of cluster assignment labels
    '''

    num_iters = 0
    converged = False

    # compute the kernel matrix if one is not provided
    if kernel_matrix is None:
        kernel_matrix = get_kernel_matrix(items, dist_func, kernel, **kernel_args)

    # initial assignment of cluster membership
    if init == "random":
        labels = __random_cluster_assignment(k, seed, items)
    else:
        labels = __proximity_cluster_assignment(k, seed, items, dist_func, init)

    item_indices_per_cluster = __get_item_indices_per_cluster(k, labels)

    while not converged:

        old_labels = deepcopy(labels)

        for item_index in range(len(items)):

            old_label = labels[item_index]
            min_distance = float("inf")
            closest_centroid = -1

            # find the nearest cluster centoid
            for centroid_index in range(k):

                dist = __distance_to_centroid(item_index, centroid_index, item_indices_per_cluster, kernel_matrix)
                if dist < min_distance:
                    min_distance = dist
                    closest_centroid = centroid_index

            # update the label of the item
            labels[item_index] = closest_centroid

            # update the item's cluster assignment
            if old_label != closest_centroid:
                item_indices_per_cluster[old_label].remove(item_index)
                item_indices_per_cluster[closest_centroid].append(item_index)


        num_iters += 1

        if labels == old_labels or num_iters >= max_iters:
            converged = True

    return labels


def get_kernel_wcss(k, items, labels, dist_func, kernel, kernel_matrix = None, **kernel_args):
    'Compute the within-cluster sum of squares value for a given label assignment'

    wcss = 0

    if kernel_matrix is None:
        kernel_matrix = get_kernel_matrix(items, dist_func, kernel, **kernel_args)

    item_indices_per_cluster = __get_item_indices_per_cluster(k, labels)

    for i in range(len(items)):
        squared_dist = __distance_to_centroid(i, labels[i], item_indices_per_cluster, kernel_matrix)
        wcss += squared_dist

    return wcss



import random
import numpy as np


def __dist_from_centroid(items, centroid_indices, dist_func):
    'Return the distance from each item to its closest centroid'

    distances = []

    for item in items:

        distances.append(min([dist_func(item, items[centroid_index]) for centroid_index in centroid_indices]))

    return np.array(distances)


def __choose_next_centroid(distances):
    'Retrun the index of the next item to become a centroid'

    probs = distances / distances.sum()
    cumprobs = probs.cumsum()
    rand = np.random.random()
    ind = np.where(cumprobs >= rand)[0][0]
    return ind


def kmeans_pp(k, seed, items, dist_func):
    'Run K-Means++ to obtain the indices of k initial centroids'

    random.seed(seed)
    np.random.seed(seed)

    centroid_indices = random.sample(range(len(items)), 1)

    while len(centroid_indices) < k:

        distances = __dist_from_centroid(items, centroid_indices, dist_func)
        next_centroid_index = __choose_next_centroid(distances)
        centroid_indices.append(next_centroid_index)

    return centroid_indices


def get_random_centroids(k, seed, items):
    'Return the indices of k random items to be used as initial centroids'

    random.seed(seed)
    return random.sample(range(len(items)), k)



import numpy as np

class Graph:
    'Class representing a graph with weighted edges and vertices'

    def __init__(self, num_vertices, edges, vertex_weights):
        self.num_vertices = num_vertices
        self.edges = edges

        # all vertices have a weight of one unless explicitly specified
        self.vertex_weights = np.ones(num_vertices)
        for node_id in vertex_weights:
            self.vertex_weights[node_id] = vertex_weights[node_id]

    def get_num_vertices(self):
        return self.num_vertices

    def get_edges(self):
        return self.edges

    def get_vertex_weights(self):
        return self.vertex_weights

    def get_adjacency_matrix(self):
        return self.adjacency_matrix

    def get_diagonal_matrix(self):
        return self.diagonal_matrix

    def compute_adjacency_matrix(self, directed = False):
        'Compute the adjacency matrix for the graph using the list of edges'

        # if the edges are undirected, two locations in the
        # adjacency matrix are modified for each edge in the list
        adjacency_matrix = np.zeros((self.num_vertices, self.num_vertices))
        for edge in self.edges:
            i, j, weight = edge
            adjacency_matrix[i, j] = weight
            if not directed:
                adjacency_matrix[j, i] = weight

        self.adjacency_matrix = adjacency_matrix

    def compute_diagonal_matrix(self):
        'Compute the diaginal matrix for the graph using its adjacency matrix'

        diagonal_matrix = np.zeros(self.adjacency_matrix.shape)
        np.fill_diagonal(diagonal_matrix, self.adjacency_matrix.sum(axis=1))
        self.diagonal_matrix = diagonal_matrix



import os
import sys
import itertools
import subprocess

import numpy as np
import networkx as nx
import matplotlib.pyplot as plt

from graph import Graph


def draw_graph(num_nodes, edges):
    'Draw a single graph'

    # create networkx graph
    G = nx.Graph()

    # add nodes
    G.add_nodes_from(range(num_nodes))

    # add edges
    G.add_edges_from(edges)

    # draw graph
    pos = nx.shell_layout(G)
    nx.draw(G, pos)
    nx.draw_networkx_labels(G, pos)

    # show graph
    plt.show()


def draw_all_graphs(file):
    'Draw all graphs in a file'

    file = open(file, 'r')

    edges = []
    num_nodes = int(file.readline())
    for line in file:

        line_elements = line.split()

        # parse more data on the current graph
        if len(line_elements) == 3:
            edge = map(lambda x: int(x), line_elements)
            edges.append(edge[:2])

        # animate the graph and clear
        elif len(line_elements) == 1:
            draw_graph(num_nodes, edges)
            edges = []
            num_nodes = int(line)

    file.close()
    draw_graph(num_nodes, edges)


def map_over_matrix_elements(matrix, function, **kwds):
    'Apply a function to every element of a matrix'

    function_vec = np.vectorize(function)
    mapped_matrix = function_vec(matrix, **kwds)

    return mapped_matrix


def draw_matrix_heat_map(matrix, color="hot"):
    'Draw a heat map of a matrix'

    plt.imshow(matrix, cmap=color, interpolation='nearest')
    plt.show()


def add_matrix_noise(matrix, percentage):
    'Append noise to the end of an n*n matrix representing n*percentage elements'

    dim_1 = int(round(percentage * matrix.shape[0]))
    dim_2 = matrix.shape[0] + dim_1

    noise = np.random.random((dim_1, dim_2)) * matrix.max()

    phase1 = np.vstack((matrix, noise[:, 0:matrix.shape[0]]))
    phase2 = np.hstack((phase1, noise.T))

    return phase2


def get_random_permutations(vals, num_permutations, seed = 0):
    'Get n permutations of the elements in vals'

    np.random.seed(seed)
    permutations = []
    all_permutations = list(itertools.permutations(vals, len(vals)))

    for index in range(num_permutations):

        permutation =  all_permutations[np.random.randint(0, len(all_permutations))]
        permutations.append(permutation)
        all_permutations.remove(permutation)

    return permutations


def load_data(file):
    'Parse a graph data file'

    graphs = []
    file = open(file, 'r')

    edges = []
    vertex_weights = {}
    num_nodes = int(file.readline())

    for line in file:

        line_elements = line.split()

        # parse vertex weight line
        if len(line_elements) == 2:
            vertex_weights[int(line_elements[0])] = float(line_elements[1])

        # parse edge line
        if len(line_elements) == 3:
            edge = map(lambda x: int(x), line_elements)
            edges.append(edge)

        # parse number of vertices line
        elif len(line_elements) == 1:
            graphs.append(Graph(num_nodes, edges, vertex_weights))
            edges = []
            vertex_weights = {}
            num_nodes = int(line)

    graphs.append(Graph(num_nodes, edges, vertex_weights))
    file.close()

    return graphs


def create_basic_edgelist_files(graphs, dir_name, common_node = False):
    'Creates files with basic edge list representations of the graphs'

    for index in range(len(graphs)):

        file = open(dir_name + '/' + "graph" + str(index) + ".txt", "w")
        graph = graphs[index]

        if common_node:

            num_vertices = graph.get_num_vertices()

            for i in range(num_vertices):
                file.write("0 " + str(i + 1) + '\n')

            for edge in graph.get_edges():
                file.write(str(edge[0]+1) + ' ' + str(edge[1] + 1) + '\n')

        else:

            for edge in graph.get_edges():
                file.write(str(edge[0]) + ' ' + str(edge[1]) + '\n')

        file.close()


def create_basic_adjacency_files(graphs, dir_name, ammend = False):
    'Creates files with basic adjacency list representations of the graphs'

    for index in range(len(graphs)):

        file = open(dir_name + '/' + "graph" + str(index) + ".txt", "w")
        graph = graphs[index]
        num_vertices = graph.get_num_vertices()

        # assumes adjacency matrix is computed
        adj_matrix = graph.get_adjacency_matrix()

        for vertex in range(num_vertices):

            file.write(str(vertex) + " ")
            row = adj_matrix[vertex]

            # there is an issue in the provided DeepWalk implementation
            # where node zero is sometimes discarded if it is disconnected
            if ammend and vertex == 0 and sum(row) == 0:
                file.write(str(1) + " ")

            for v in range(len(row)):

                if row[v] != 0:
                    file.write(str(v) + " ")

            # do not print a new line after the final row
            if vertex != num_vertices - 1:
                file.write("\n")

        file.close()


def load_matrix_data(file, lines_per_matrix):
    'Load a file with concatenated matrix data'

    matrices = []
    file = open(file, 'r')

    matrix = []
    num_lines = 0

    for line in file:

        row = [float(num) for num in line.split()]
        matrix.append(row)
        num_lines += 1

        if num_lines == lines_per_matrix:

            matrices.append(np.array(matrix))
            num_lines = 0
            matrix = []

    file.close()

    return matrices


def load_deep_walk_files(dir_name, lines_per_matrix):
    'Load all DeepWalk formated files from a directory'

    matrices = []
    num_files = len(os.listdir(dir_name))

    for index in range(num_files):

        file = open(dir_name + '/' + "graph" + str(index) + ".txt", "r")
        adj_dict = {}
        matrix = []

        # discard the first line of each file - unneeded metadata
        line = file.readline()

        for line in file:
            row = [float(num) for num in line.split()]
            adj_dict[row[0]] = row[1:]

        for i in range(lines_per_matrix):
            matrix.append(adj_dict[i])

        matrices.append(np.array(matrix))
        adj_dict = {}
        matrix = []

        file.close()

    return matrices


def get_rolx_matrices(graphs, rolx_path = "/path/to/rolx", num_roles = 3):
    'Convert a collection of graphs into RolX matrices'

    # the matrix has num_node lines + the extra node connected to everything
    lines_per_matrix = graphs[0].get_num_vertices() + 1
    
    # create directoriries where the intermediate files will be stored
    subprocess.call(["mkdir", "format_conversion"])
    subprocess.call(["mkdir", "format_conversion/in"])
    subprocess.call(["mkdir", "format_conversion/out"])

    # create the edge lists which will be provided to RolX
    create_basic_edgelist_files(graphs, "format_conversion/in", common_node = True)

    input_file_names = ["format_conversion/in/graph" + str(i) + ".txt" for i in range(len(graphs))]
    output_file_names = ["format_conversion/out/graph" + str(i) + ".txt" for i in range(len(graphs))]

    # conver each graph into a RolX matrix
    for i in range(len(graphs)):

        # this method assumes that RolX has been modified to putput the NxR matrix
        rolx_cmd = [rolx_path, "-i:" + input_file_names[i], "-o:" + output_file_names[i], "-l:" + str(num_roles), "-u:" + str(num_roles)]
        subprocess.call(rolx_cmd)

    # concatenate all output matrix files into a single file
    cat_cmd = ["cat"] + output_file_names
    with open("format_conversion/out/graphs.txt", "w") as outfile:
        subprocess.call(cat_cmd, stdout=outfile)

    # read in the RolX matrices
    matrices = load_matrix_data("format_conversion/out/graphs.txt", lines_per_matrix)

    # discard the common node from each output matrix
    matrices = [m[1:] for m in matrices]

    # delete the intermediate files
    subprocess.call(["rm", "-rf", "format_conversion"])

    return matrices


def get_deep_walk_matrices(graphs, representation_size = 64, number_walks = 10, walk_length = 40, undirected = True):
    'Convert a collection of graphs into DeepWalk matrices'

    # the matrix has num_node lines
    lines_per_matrix = graphs[0].get_num_vertices()

    # create directoriries where the intermediate files will be stored
    subprocess.call(["mkdir", "format_conversion"])
    subprocess.call(["mkdir", "format_conversion/in"])
    subprocess.call(["mkdir", "format_conversion/out"])

    # create the adjacency lists which will be provided to DeepWalk
    create_basic_adjacency_files(graphs, "format_conversion/in", ammend = True)

    input_file_names = ["format_conversion/in/graph" + str(i) + ".txt" for i in range(len(graphs))]
    output_file_names = ["format_conversion/out/graph" + str(i) + ".txt" for i in range(len(graphs))]

    # conver each graph into a DeepWalk matrix
    for i in range(len(graphs)):

        # this method assumes that DeepWalk has been insatlled
        dw_cmd = ["deepwalk", "--input", input_file_names[i], "--output", output_file_names[i], 
                "--representation-size", str(representation_size), "--number-walks", str(number_walks),
                "--walk-length", str(walk_length), "--undirected", str(undirected)]
        subprocess.call(dw_cmd)

    # read in the DeepWalk matrices
    matrices = load_deep_walk_files("format_conversion/out", lines_per_matrix)

    # delete the intermediate files
    subprocess.call(["rm", "-rf", "format_conversion"])

    return matrices



import itertools
import numpy as np


def __assert_all_nodes_in_regions(num_of_nodes, regions):
    'Assert that every oned is included in exactly one region'

    all_nodes = range(num_of_nodes)
    nodes_in_region = []
    for region in regions:
        nodes_in_region += range(region[0], region[1])
    nodes_in_region.sort()

    return nodes_in_region == all_nodes


def __get_num_of_edge_in_region(num_of_nodes, directed, percentage):
    'The number of edges will be equal to a percentage of the edges in a clique'

    edges_in_clique = (num_of_nodes * (num_of_nodes - 1)) / 2
    num_of_edges = int(round(edges_in_clique * percentage))

    # if the edges are directed we can have twice as many
    if directed:
        num_of_edges = num_of_edges * 2

    return num_of_edges


def __generate_edge_weight(weighted_edges, edge_weight_range):
    'Generate an edge weight'

    # if the edge is unweighted assign a weight of one, otherwise assign a random weight in the given range
    if not weighted_edges:
        weight = 1
    else:
        weight = np.random.randint(edge_weight_range[0], edge_weight_range[1] + 1)

    return weight


def __generate_edge_between_regions(edges, directed, region_0, region_1):
    'Create a single edge between the two regions'

    approved = False

    while not approved:

        node1 = np.random.randint(region_0[0], region_0[1])
        node2 = np.random.randint(region_1[0], region_1[1])
        
        if not directed:
            edge = [node1, node2]
            edge.sort()
        else:
            # randomly choose the source/destination region
            if np.random.random() > 0.5:
                edge = [node1, node2]
            else:
                edge = [node2, node1]

        if edge not in edges:
            edges.append(edge)
            approved = True


def __generate_uniform_cross_region_edges(edges, directed, region_con, regions):
    'Create an equal number of edges joining all region pairs'

    region_pairs = list(itertools.combinations(regions, 2))

    for region_pair in region_pairs:

        region_0 = region_pair[0]
        region_1 = region_pair[1]

        smaller_group_size = min(len(range(region_0[0], region_0[1])), len(range(region_1[0], region_1[1])))
        num_edges_between_regions = int(round(smaller_group_size * region_con))

        for e in range(num_edges_between_regions):
            __generate_edge_between_regions(edges, directed, region_0, region_1)




def __generate_cross_region_edges(edges, directed, num_of_nodes, region_con, regions):
    'Randomly create edges joining region pairs'

    region_pairs = list(itertools.combinations(regions, 2))
    avg_region_size = num_of_nodes / float(len(regions))
    num_edges_between_regions = int(round(region_con * avg_region_size * len(region_pairs)))

    for e in range(num_edges_between_regions):

        region_pair = region_pairs[np.random.randint(0, len(region_pairs))]
        __generate_edge_between_regions(edges, directed, region_pair[0], region_pair[1])




def __generate_edges(num_of_nodes, weighted_edges, edge_weight_range, directed, regions, region_con, uniform_region_con):
    'Generate all edges for a given graph'

    edges = []

    # generate edges within each region
    for region in regions:

        num_nodes_in_region = len(range(region[0], region[1]))
        num_edges_in_region = __get_num_of_edge_in_region(num_nodes_in_region, directed, region[2])

        for e in range(num_edges_in_region):

            approved = False

            while not approved:

                node1 = np.random.randint(region[0], region[1])
                node2 = np.random.randint(region[0], region[1])
                edge = [node1, node2]

                if not directed:
                    edge.sort()

                # currently excluding self edges
                if edge[0] != edge[1] and edge not in edges:
                    edges.append(edge)
                    approved = True

    # generate edges between different regions
    if uniform_region_con:
        __generate_uniform_cross_region_edges(edges, directed, region_con, regions)
    else:
        __generate_cross_region_edges(edges, directed, num_of_nodes, region_con, regions)

    # include a weight for each edge
    for edge in edges:
        weight = __generate_edge_weight(weighted_edges, edge_weight_range)
        edge.append(weight)

    return edges


def __generate_node_weights(regions, node_weight_ranges):
    'For each node, produce a weight in the specified weight range for its region'

    node_weights = []
    node_ids_in_regions = [range(region[0], region[1]) for region in regions]

    for region_id in range(len(regions)):

        node_ids_in_region = node_ids_in_regions[region_id]
        node_weight_range = node_weight_ranges[region_id]

        for node_id in node_ids_in_region:

            node_weight = (node_weight_range[1] - node_weight_range[0]) * np.random.random() + node_weight_range[0]
            node_weights.append([node_id, node_weight])

    return node_weights


def generate_graphs_file_2(file_name, num_of_graphs, num_of_nodes, seed = 0,
                         weighted_edges = False, edge_weight_range = [1, 10], directed = False,
                         regions = [], region_con = 0.1, uniform_region_con = True,
                         weighted_nodes = False, node_weight_ranges = []):

    """Generate a collection of graphs with a set of regions with variable connectivity

    Parameters
    ----------
    file_name: string
        name of the generated graphs file

    num_of_graphs: int
        the nuber of graphs that will be generated and written into the file

    num_of_nodes: int
        the number of nodes that every graph will have

    seed: int
        a seed paramether for controlling randomness

    weighted_edges: boolean
        determines if the edges will be weighted or not

    edge_weight_range: [lwb: int, upb: int]
        the weight of each edge will be a random integer form lwb to upb

    directed: boolean
        determines if the edges will be directed or not

    regions: list of (region_start_node: int, region_end_node: int, region_connectivity: float) tuples
        outlines the partitioning of the graph into regions along with their connectivity level
        (percentage of the edges in a clique made up of the nodes in the region)

    region_con: float
        determines tha cross region connectivity

    uniform_region_con: boolean
        determines if the number of edges between each pair of regions are equal

    weighted_nodes: boolean
        determines if the nodes will be weighted or not

    node_weight_ranges: list of (lwb: float, upb: float) tuples
        the lwb and upb of node weights for each region
    """
    
    # assert that the input parameters provide consistent information
    if not __assert_all_nodes_in_regions(num_of_nodes, regions):
        print("Some nodes are not part of a region or are included in multiple regions")
        return

    if weighted_nodes and len(regions) != len(node_weight_ranges):
        print("The number of node regions and node weight ranges do not match")
        return

    np.random.seed(seed)
    file = open(file_name, 'w')

    # for each graph write out - number of nodes, node weights, edges
    for graph in range(num_of_graphs):

        # write out the number of nodes in the graph - also indicates the start of a new graph
        file.write(str(num_of_nodes) + '\n')

        # write out all node weights
        if weighted_nodes:
            for node_weight in __generate_node_weights(regions, node_weight_ranges):
                file.write(str(node_weight[0]) + ' ' + str(node_weight[1]) + '\n')

        # write out all edges in the graph
        edges = __generate_edges(num_of_nodes, weighted_edges, edge_weight_range,
                                directed, regions, region_con, uniform_region_con)

        for edge in edges:
            file.write(str(edge[0]) + ' ' + str(edge[1]) + ' ' + str(edge[2]) + '\n')

    file.close()



import numpy as np
from kmeans_init import get_random_centroids, kmeans_pp


def __compute_cluster_assignment(centroid_indices, items, distance_matrix):
    'Get a dictionary with cluster ids as keys and arrays of assinged item indices as values'

    labels = {}

    for centroid_index in centroid_indices:
        labels[centroid_index] = []

    for item_index in range(len(items)):
        
        min_distance = float("inf")
        closest_centroid = -1

        for centroid_index in centroid_indices:

            dist = distance_matrix[centroid_index, item_index]

            if dist < min_distance:
                min_distance = dist
                closest_centroid = centroid_index

        labels[closest_centroid].append(item_index)

    return labels


def __compute_new_cluster_centroids(labels, items, distance_matrix):
    'The new centroids are the items with minimum average distance to all other items in their cluster'

    new_centroid_indices = []

    for old_centroid_index in labels:

        cluster = labels[old_centroid_index]

        min_average_dist = float("inf")
        new_centroid = -1

        for item_index in cluster:

            distance = 0
            for i_i in cluster:
                distance += distance_matrix[item_index, i_i]
            
            average_dist = distance / len(cluster)

            if average_dist < min_average_dist:
                min_average_dist = average_dist
                new_centroid = item_index

        new_centroid_indices.append(new_centroid)

    return new_centroid_indices


def get_distance_matrix(items, dist_func):
    'Compute the NxN matrix of distances between all pairs of items'

    distance_matrix = np.zeros((len(items), len(items)))

    j = 0
    for i in range(len(items)):

        while j < len(items):

            distance = dist_func(items[i], items[j])
            distance_matrix[i, j] = distance
            distance_matrix[j, i] = distance

            j += 1

        j = i + 1

    return distance_matrix


def sd_kmeans(k, max_iters, seed, items, dist_func, init = "random", distance_matrix = None):
    """Run Structured Data K-Means

    Parameters
    ----------
    k: int
        the number of clusters

    max_iters: int
        the number of maximum allowed itterations before the method terminates
        (termination will occure earlier if the chosen centroids converge)

    seed: int
        a seed paramether for controlling randomness

    items: 1-D list
        the list of structured data items thet will be clustered

    dist_func: function pointer
        a function for computing the distance between any two items in the items list

    init: string
        a string indicating the centroid initialization method

    distance_matrix: 2-D numpy array
        an optional pre-computed NxN matrix of distance values between all pairs of items

    The possible initialization methods are: 'random' and 'kpp':
        random - randomly choose centroids amongst the input items
        kpp    - use k-means++ to choose centroids amongst the input items

    Return
    ------
    labels: 1-D list of ints
        the indices of the input items which best approximate the centroids
    """

    num_iters = 0
    converged = False

    # compute the distance matrix if one is not provided
    if distance_matrix is None:
        distance_matrix = get_distance_matrix(items, dist_func)

    # select k distinct items as the cluster centers
    if init == "random":
        centroid_indices = get_random_centroids(k, seed, items)
    else:
        centroid_indices = kmeans_pp(k, seed, items, dist_func)

    while not converged:

        labels = __compute_cluster_assignment(centroid_indices, items, distance_matrix)

        new_centroid_indices = __compute_new_cluster_centroids(labels, items, distance_matrix)

        num_iters += 1

        if new_centroid_indices == centroid_indices or num_iters >= max_iters:
            converged = True
        else:
            centroid_indices = new_centroid_indices

    return centroid_indices


def get_sd_labels(centroid_indices, items, dist_func, distance_matrix = None):
    'Compute the assigned cluster label for each item given the indices of the approximate centroids'

    labels = []

    for item_index in range(len(items)):
        
        min_distance = float("inf")
        label = -1

        # the distance matrix is not computed here as n*n >> n*k
        for index in range(len(centroid_indices)):

            # compute or look up the distance
            if distance_matrix is None:
                dist = dist_func(items[centroid_indices[index]], items[item_index])
            else:
                dist = distance_matrix[centroid_indices[index], item_index]

            if dist < min_distance:
                min_distance = dist
                label = index

        labels.append(label)

    return labels


def get_sd_wcss(centroid_indices, items, dist_func, distance_matrix = None):
    'Compute the within-cluster sum of squares value for a set of approximate centroids'

    wcss = 0
    labels = get_sd_labels(centroid_indices, items, dist_func, distance_matrix)

    for item_index in range(len(labels)):
        centroid_index = centroid_indices[labels[item_index]]

        # compute or look up the distance
        if distance_matrix is None:
            dist = dist_func(items[item_index], items[centroid_index])
        else:
            dist = distance_matrix[item_index, centroid_index]

        wcss += (dist ** 2)

    return wcss



from setuptools import setup, find_packages

setup(name='graphSimilarity',
      version='0.1.0',
      description='Graph Distance Measures and Clustering',
      author='Ivan Kyosev',
      author_email='2092065k@student.gla.ac.uk',
      license='GPL',
      install_requires=['numpy',
                        'matplotlib',
                        'networkx',
                        'scipy',
                        'scikit_learn'
                        ],
      package_data={'graphSimilarity': ['README.md']},
      packages=find_packages())
